from __future__ import absolute_import
from ._version import __version__

__version__  # silence codequality

from .membership import RolesNode
from . import permissions  # nopep8
from .permissions import requires

SetMembershipNode = RolesNode  # For backwards compatibility purposes.
requires



from collections import OrderedDict
from operator import or_

from django.conf.urls import url
from django.conf.urls import include
from django.contrib.admin.sites import AdminSite

from baya.permissions import requires
from baya.membership import ValueNode
from functools import reduce


# Keep a registry of baya-enabled admin sites so we can properly intercept
# permissions checking in NestedLDAPGroupsBackend.
_admin_registry = set()


class NestedGroupsAdminSite(AdminSite):
    def __init__(self, *args, **kwargs):
        super(NestedGroupsAdminSite, self).__init__(*args, **kwargs)
        _admin_registry.add(self)
        all_groups = self._get_required_baya_groups()
        if all_groups is not None:
            self.index = requires(all_groups)(self.index)
            self.app_index = requires(all_groups)(self.app_index)

    def _get_admins_with_gate(self):
        admins = []
        # Only need to sort this list so we get consistent OperatorNode trees.
        items = sorted(self._registry.items(),
                       key=lambda x: x[0]._meta.model_name)
        for model, model_admin in items:
            if hasattr(model_admin, '_gate'):
                admins.append((model, model_admin))
        return admins

    def get_urls(self):
        """Ensure that urls included in get_urls() are behind requires().

        We need to fix the include() logic for admin URLs. Django admin isn't
        very extensible, so we have to call super, remove the url patterns
        for model_admins that have a _gate, and replace the pattern with
        a properly built include behind the model admin's gate.

        Would be a lot easier if django exposed something like
        get_patterns_for_app(app_label), but noooooo.
        """
        # We have to maintain the URL ordering due to the way URLs are resolved
        # TODO - Test this, can lead to heisenbugs
        urls = OrderedDict((urlp.regex.pattern, urlp) for urlp in
                           super(NestedGroupsAdminSite, self).get_urls())
        for model, model_admin in self._get_admins_with_gate():
            if hasattr(model._meta, 'module_name'):
                model_name = model._meta.module_name
            elif hasattr(model._meta, 'model_name'):
                model_name = model._meta.model_name
            else:
                raise ValueError(
                    "Model Admin is missing a module or model name.")
            pattern = (
                r'^%s/%s/' %
                (model._meta.app_label, model_name))
            urls[pattern] = url(
                pattern,
                requires(get=model_admin._gate.get_requires,
                         post=model_admin._gate.post_requires)(
                             include(model_admin.urls)))
        return list(urls.values())

    def _get_required_baya_groups(self, app_label=None):
        # Loop over all model admins, checking their set of permissions
        all_roles = OrderedDict()
        for model, model_admin in self._get_admins_with_gate():
            if app_label and app_label != model._meta.app_label:
                continue
            roles = model_admin._gate.get_requires
            if roles is not None and not isinstance(roles, ValueNode):
                all_roles[roles] = True
        if all_roles:
            return reduce(or_, all_roles.keys())
        return None

    def admin_view(self, view, cacheable=False):
        fn = super(NestedGroupsAdminSite, self).admin_view(view, cacheable)
        all_groups = self._get_required_baya_groups()
        if all_groups is not None:
            fn = requires(all_groups)(fn)
        return fn



"""
Visitor pattern for walking through a BaseNode AST.
"""

from collections import namedtuple

from .membership import AndNode
from .membership import OperatorNode
from .membership import OrNode
from .membership import RolesNode
from .membership import ValueNode
from .membership import XorNode


VisitedNode = namedtuple('VisitedNode', ['node', 'value'])


class NodeVisitor(object):

    def visit(self, node, **kwargs):
        """
        Visitor method dispatcher based on node type or operator node arity.

        Input: BaseNode
        """
        if isinstance(node, RolesNode):
            return self._visit_roles_node(node, **kwargs)

        if isinstance(node, ValueNode):
            return self._visit_value_node(node, **kwargs)

        if isinstance(node, OperatorNode):
            return self._visit_operator_node(node, **kwargs)

        raise TypeError('Cannot visit node %r' % node)

    def _visit_value_node(self, value_node, **kwargs):
        """Return the value from a ValueNode."""
        return value_node.value

    def _visit_roles_node(self, roles_node, **kwargs):
        """
        Abstract visit method for the leaf node type RolesNode.

        Input: RolesNode
        """

        raise NotImplementedError(
            'Child classes of NodeVisitor must implement _visit_roles_node.')

    def _visit_operator_node(self, operator_node, **kwargs):
        """
        Dispatcher method for OperatorNode types.

        Currently dispatches based on the arity of the operator.

        Input: OperatorNode
        """

        visited_operands = [VisitedNode(operand, self.visit(operand, **kwargs))
                            for operand in operator_node._operands]
        dispatch_methods = [
            self._visit_nullary_node,
            self._visit_unary_node,
            self._visit_binary_node,
        ]
        return dispatch_methods[operator_node.arity](operator_node,
                                                     *visited_operands)

    def _visit_nullary_node(self, operator_node):
        raise ValueError('There are no nullary operators yet.')

    def _visit_unary_node(self, operator_node, visited_operand):
        raise NotImplementedError(
            'Child classes of NodeVisitor must implement _visit_unary_node.')

    def _visit_binary_node(self, operator_node, left_visited_operand,
                           right_visited_operand):
        raise NotImplementedError(
            'Child classes of NodeVisitor must implement _visit_binary_node.')


class ExpressionWriter(NodeVisitor):
    """
    NodeVisitor concrete class that writes out a pretty-printed expression.

    For example, instead of naively interpreting the following node structure
    > OrNode(XorNode('A', 'B'), XorNode('C', 'D'))
    as '((A ^ B) | (C ^ D))' it will return 'A ^ B | C ^ D'.
    But for the following node structure
    > AndNode(OrNode('A', 'B'), OrNode('C', 'D'))
    it will still print it as '(A | B) & (C | D)'.

    All concrete visit methods will return strings.
    """

    # Items that appear first have higher precedence.
    binary_operator_precedence = [
        AndNode,
        XorNode,
        OrNode,
    ]

    def _visit_value_node(self, value_node, **kwargs):
        return str(super(ExpressionWriter, self)._visit_value_node(
            value_node, **kwargs))

    def _visit_roles_node(self, roles_node, **kwargs):
        return str(roles_node)

    def _visit_unary_node(self, operator_node, visited_operand):
        args = (operator_node.display_name, visited_operand.value)
        arity = self._get_arity(visited_operand.node)
        if arity is not None and arity > 1:
            return '%s(%s)' % args
        return '%s%s' % args

    def _visit_binary_node(self, operator_node, left_visited_operand,
                           right_visited_operand):
        def get_operand_value(visited_operand):
            value = visited_operand.value
            if (self._get_arity(visited_operand.node) == 2 and
                    self._has_precedence_over(operator_node,
                                              visited_operand.node)):
                value = '(%s)' % value
            return value

        return '%s %s %s' % (get_operand_value(left_visited_operand),
                             operator_node.display_name,
                             get_operand_value(right_visited_operand))

    @staticmethod
    def _get_arity(node):
        return getattr(node, 'arity', None)

    @classmethod
    def _has_precedence_over(cls, left_node, right_node):
        """
        Helper method that determines which operator node has precedence.

        Returns True if left_node has strict precedence over right_node.
        """

        return (cls.binary_operator_precedence.index(left_node.__class__) <
                cls.binary_operator_precedence.index(right_node.__class__))


class PermissionChecker(NodeVisitor):
    """
    NodeVisitor concrete class that checks whether the node has permissions.

    PermissionChecker is instantiated with a roles set, which is an iterable of
    strings which represent role names.  This visitor then evaluates the
    expression encoded in the node AST.  Each node is an operator or a leaf
    node.  Currently leaf nodes are RolesNodes, which evaluate to True if and
    only if their roles set is a subset of the PermissionChecker's roles set.
    Then the operator nodes perform operations on the child booleans and
    propagate the final value up to the root.

    All concrete visit methods will return a boolean.
    """

    def __init__(self, roles):
        """Instantiate a PermissionChecker.

        Args:
            roles: An iterable of strings, representing roles a user has.

        Usage:

            from baya.utils import group_names
            from baya.membership import RolesNode as g

            user_groups = group_names(request.user.ldap_user.group_dns)
            # user_groups looks like {'group1', 'group2', ...}
            checker = PermissionChecker(user_groups)
            required_groups = g('req1') & g('req2')
            user_has_permissions = checker.visit(required_groups)
        """
        self._roles_set = {role.lower() for role in roles}

    def _visit_roles_node(self, roles_node, **kwargs):
        return roles_node.get_roles_set(**kwargs) <= self._roles_set

    def _visit_unary_node(self, operator_node, visited_operand):
        return operator_node.operator(visited_operand.value)

    def _visit_binary_node(self, operator_node, left_visited_operand,
                           right_visited_operand):
        return operator_node.operator(left_visited_operand.value,
                                      right_visited_operand.value)



def permission_context(request):
    func = request.resolver_match.func
    gate = getattr(func, '_gate', None)
    data = {}
    if gate:
        data = {
            'baya_permissions': gate,
            'baya_membership': gate.get_membership_node(request),
        }

    if hasattr(request, 'baya_permission_denied'):
        data['baya_permission_denied'] = request.baya_permission_denied
    return data



from baya.utils import has_permission
from django import template
from django.core.urlresolvers import resolve
from django.core.urlresolvers import reverse

register = template.Library()


@register.assignment_tag(takes_context=True)
def can_user_perform_action(context, action, *args, **kwargs):
    """
    Assignment tag to check user permission within a template.

    Example:
        {%  can_user_perform_action "home" as can_view_homepage %}

    Args:
        context: The template context (implicitly passed in because
            takes_context=True)
        action: The name of the url
        args/kwargs: The args/kwargs required by reverse

    Returns:
        bool: True if user has permission, False otherwise.

    Caveats:
        If there is no Gate (no requires function wrapping the viewfunc),
        has_permission returns False.

        action, args, and kwargs are fed directly into reverse. If they aren't
        given correctly, exceptions will be thrown. e.g. You supply both
        args and kwargs. For details please see django docs:
        https://docs.djangoproject.com/en/1.8/ref/urlresolvers/#reverse
    """
    view_func = resolve(reverse(action, args=args, kwargs=kwargs)).func

    return has_permission(view_func, context['user'], 'any')



import collections
import functools

import django
from django.conf import settings
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.admin.options import InlineModelAdmin
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import RegexURLPattern
from django.core.urlresolvers import RegexURLResolver
import six

from .membership import BaseNode
from .membership import ValueNode
from .membership import RolesNode
from .utils import group_names
from .utils import user_in_group
from .visitors import ExpressionWriter


DENY_ALL = ValueNode(False)
ALLOW_ALL = ValueNode(True)


class Gate(object):
    """Track the groups that a view requires.

    The `requires` decorator attaches an instance of this class to every
    view it decorates.
    """
    GET_METHODS = {'GET', 'HEAD', 'OPTIONS', 'TRACE'}
    POST_METHODS = {'POST', 'PUT', 'PATCH', 'DELETE', 'CONNECT'}
    DEFAULT_PERMISSION_NODE = RolesNode

    def __init__(self, all_requires=None,
                 get_requires=None, post_requires=None,
                 login_url=None):
        """Create a Gate object, which protects access to views.

        Args:
            all_requires: Groups which are required for access via any HTTP
                verb.
            get_requires: Groups which are required for GET-like access. See
                Gate.GET_METHODS
            post_requires: Groups which are required for POST-like access. See
                Gate.POST_METHODS
            login_url: The URL to redirect to when the user is unauthenticated.
                Defaults to django's LOGIN_URL setting. You can set a
                baya-specific default by setting BAYA_LOGIN_URL in your
                django settings.
        """
        baya_login_url = getattr(settings, 'BAYA_LOGIN_URL', None)
        if login_url is not None:
            self.login_url = login_url
        elif baya_login_url is not None:
            self.login_url = baya_login_url
        else:
            self.login_url = settings.LOGIN_URL

        if all_requires is None:
            self.get_requires = self._ensure_permission_node(get_requires)
            self.post_requires = self._ensure_permission_node(post_requires)
        else:
            all_requires = self._ensure_permission_node(all_requires)
            self.get_requires = all_requires
            self.post_requires = all_requires
            if get_requires is not None:
                self.get_requires &= self._ensure_permission_node(get_requires)
            if post_requires is not None:
                self.post_requires &= self._ensure_permission_node(
                    post_requires)

    def _ensure_permission_node(self, groups):
        """
        Convert a 'groups' parameter (from __init__) into a PermissionNode.
        """
        if isinstance(groups, BaseNode):
            return groups

        if groups is None:
            groups = []
        elif isinstance(groups, six.string_types):
            groups = [group.strip() for group in groups.split(',')]

        return self.DEFAULT_PERMISSION_NODE(*groups)

    def _has_permission(self, user, membership_node, request=None):
        return user_in_group(user, membership_node, request=request)

    def has_get_permission(self, request):
        return self._has_permission(request.user, self.get_requires, request)

    def user_has_get_permission(self, user):
        return self._has_permission(user, self.get_requires)

    def has_post_permission(self, request):
        return self._has_permission(request.user, self.post_requires, request)

    def user_has_post_permission(self, user):
        return self._has_permission(user, self.post_requires)

    def has_any_permission(self, request):
        return (self.has_get_permission(request) or
                self.has_post_permission(request))

    def user_has_any_permission(self, user):
        return (self.user_has_get_permission(user) or
                self.user_has_post_permission(user))

    def get_membership_node(self, request):
        if request.method in self.GET_METHODS:
            return self.get_requires
        elif request.method in self.POST_METHODS:
            return self.post_requires

    def get_permissions_required_data(self, request):
        user_groups = []
        user_groups_str = "{}"
        if hasattr(request, 'user') and hasattr(request.user, 'ldap_user'):
            user_groups = sorted(
                group_names(request.user.ldap_user.group_dns))
            user_groups_str = "{%s}" % ", ".join(
                str(el) for el in user_groups)
        data = {
            'requires_groups': self.get_membership_node(request),
            'requires_groups_str': ExpressionWriter().visit(
                self.get_membership_node(request)),
            'user_groups': user_groups,
            'user_groups_str': user_groups_str,
        }
        return data

    def allow_or_deny(self, request):
        """
        Allow the request to continue (return None), redirect to login,
        or raise PermissionDenied.

        These weird semantics are so you can do:
        return allow_or_deny(request) or HttpResponse(...)
        """
        from django.contrib.auth.views import redirect_to_login
        # Annotate request with some baya information
        request.baya_requires = self.get_permissions_required_data(request)
        # Set BAYA_ALLOW_ALL while testing in development to disable
        # permissions checking.
        if getattr(settings, 'BAYA_ALLOW_ALL', False):
            return None

        # Check permissions before checking if the user is authenticated,
        # to allow views to be protected by empty RolesNodes to be served.
        if (request.method in self.POST_METHODS and
                self.has_post_permission(request)):
            return None
        elif (request.method in self.GET_METHODS and
                self.has_get_permission(request)):
            return None

        is_authenticated = (
            request.user.is_authenticated() if django.VERSION[:2] < (1, 10)
            else request.user.is_authenticated)
        if not is_authenticated:
            path = request.get_full_path()
            return redirect_to_login(path, self.login_url)

        permission_denied_msg = (
            "User {user} does not have permission to {method} to this "
            "resource. Groups {groups} are required, but {user} only has "
            "{user_groups}").format(
                user=request.user,
                method=request.method,
                groups=request.baya_requires['requires_groups'],
                user_groups=request.baya_requires['user_groups'])
        raise PermissionDenied(permission_denied_msg)

    def __iadd__(self, other):
        self.get_requires &= other.get_requires
        self.post_requires &= other.post_requires
        # Prefer other's login_url, if set
        if (other.login_url is not None and
                six.text_type(other.login_url) != six.text_type(settings.BAYA_LOGIN_URL)):  # nopep8
            self.login_url = other.login_url
        return self

    def __add__(self, other):
        new = self.__class__()
        new.get_requires = self.get_requires
        new.post_requires = self.post_requires
        new.login_url = self.login_url
        new += other
        return new

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.get_requires == other.get_requires and
                self.post_requires == other.post_requires and
                self.login_url == other.login_url)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "Gate(get_requires=%s, post_requires=%s)" % (
            repr(self.get_requires), repr(self.post_requires))


class requires(object):
    """Decorate methods or urls that need to be access controlled.

    The requires decorator takes a BaseNode as parameters. You can
    read more about those in the baya.membership module. You really just
    need to know about RolesNode, which is usally imported as 'g':

        from baya import RolesNode as g

    Examples:

    Both GET and POST need 'group1' and 'group2' permissions:

        @requires(g('group1') & g('group2'))
        def my_view(request):
            ...

    Both GET and POST require one permission, and POST requires an extra.
    The first parameter is &-ed with the following get and post parameters:

        @requires(g('both_methods'), post=g('and_another_method'))
        def my_view(request):
            ...

    Specify get and post required permissions individually:

        @requires(get=g('group1') | g('group2'), post=g('group2'))
        def my_view(request):
            ...

    Protect a url:

        url(r'^protected_url/$', requires(g('admin'))(my_view))

    Protect an entire url include. Note that when you nest requires() calls
    you are &-ing them together.

        url(r'^admin_views/', requires(g('admin'))(include(my_admin.urls)))

    All of the binary logic operations are available: & | ^ ~.
    """
    GATE_MODEL = Gate

    def __init__(self, groups=None, get=None, post=None, **kwargs):
        self.gate = self.GATE_MODEL(
            all_requires=groups,
            get_requires=get,
            post_requires=post,
            login_url=kwargs.get('login_url'))

    def decorate_method(self, fn, *args, **kwargs):
        """Decorate a view method.

        This is pretty simple - just attach a Gate as a `_gate` property
        to the function and then wrap it in the method dispatcher.
        """
        @functools.wraps(fn)
        def dispatcher(*args, **kwargs):
            largs = list(args)
            if isinstance(largs[0], BaseModelAdmin):
                func = functools.partial(fn, largs.pop(0))
            else:
                func = fn
            request = largs.pop(0)
            return (self.gate.allow_or_deny(request) or
                    func(request, *largs, **kwargs))
        dispatcher._gate = self.gate
        return dispatcher

    def decorate_url_pattern(self, pattern, *args, **kwargs):
        """Decorate a RegexURLPattern or RegexURLResolver.

        Args:
            resolve_fn: Either RegexURLPattern or RegexURLResolver, from
                django.core.urlresolvers

        This decorates the callback for a url after it gets resolved with
        self.decorate_method.
        """
        resolve_fn = pattern.resolve

        @functools.wraps(resolve_fn)
        def patch_resolve(path):
            result = resolve_fn(path)
            if result:
                result.func = self.decorate_method(
                    result.func, *args, **kwargs)
            return result
        pattern.resolve = patch_resolve
        return pattern

    def decorate_admin(self, cls, *args, **kwargs):
        """Alter an admin view for access control.

        This doesn't actually decorate the given class, but instead decorates
        change_view and changelist_view. This is necessary because we often
        want to allow GET access to the chagelist/change_view but not POST,
        since those views are often used as information sources.

        Admin views also need to be in a admin.sites.NestedGroupsAdminSite
        again because we need to properly decorate the admin urls, which are
        provided via NestedGroupsAdminSite.get_urls().
        """
        if not hasattr(cls, '_gate'):
            setattr(cls, '_gate', self.gate)
        else:
            cls._gate += self.gate
            self.gate = cls._gate

        cls.add_view = self.decorate_method(
            cls.add_view, *args, **kwargs)
        cls.changelist_view = self.decorate_method(
            cls.changelist_view, *args, **kwargs)
        cls.change_view = self.decorate_method(
            cls.change_view, *args, **kwargs)
        cls.delete_view = self.decorate_method(
            cls.delete_view, *args, **kwargs)
        return cls

    def decorate_include(self, inc_tup, *args, **kwargs):
        """Decorate a url include with a requires().

        Args:
            inc_tuple: The tuple returned from executing include(my_app.urls).

        Returns the modified include tuple. Note that the url-patterns are
        modified in-place.
        """
        if isinstance(inc_tup[0], (list, tuple)):
            # Must be an include from the admin, where instaed of setting
            # a single module, it returns a list(?!) of patterns...
            patterns = inc_tup[0]
        elif hasattr(inc_tup[0], 'urlpatterns'):
            patterns = inc_tup[0].urlpatterns
        else:
            raise TypeError("Unknown include type: %s" % repr(inc_tup))

        for pattern in patterns:
            self.decorate_url_pattern(pattern, *args, **kwargs)
        return inc_tup

    def __call__(self, fn, *args, **kwargs):
        """Delegate the decoration to the appropriate method."""
        if isinstance(fn, functools.partial) and not hasattr(fn, '__module__'):
            raise ValueError(
                'Cannot decorate a bare functools.partial view.  '
                'You must invoke functools.update_wrapper(partial_view, '
                'full_view) first.')
        if not isinstance(fn, type) and isinstance(fn, collections.Callable):
            return self.decorate_method(fn, *args, **kwargs)
        elif isinstance(fn, tuple):
            # Must be an include('my_app.urls') we're decorating
            return self.decorate_include(fn, *args, **kwargs)
        elif isinstance(fn, (RegexURLPattern, RegexURLResolver)):
            return self.decorate_url_pattern(fn, *args, **kwargs)
        elif isinstance(fn, type) and issubclass(fn, BaseModelAdmin):
            if issubclass(fn, InlineModelAdmin):
                raise TypeError("Cannot decorate Inlines. See "
                                "baya.admin.options.BayaInline instead.")
            return self.decorate_admin(fn, *args, **kwargs)
        elif isinstance(fn, six.string_types):
            raise TypeError("Cannot decorate string-path to view: %s." % fn)
        else:
            # You'll probably only get here if you're trying to decorate
            # a class-based view
            raise TypeError(
                "Invalid type for requires decorator (%s). "
                "You cannot decorate class-based views. Decorate "
                "the URL or the as_view method instead." % type(fn))



from functools import partial
import six

from django.contrib.auth import get_permission_codename
from django_auth_ldap.backend import LDAPBackend


class NestedLDAPGroupsBackend(LDAPBackend):
    def get_all_permissions(self, user, obj=None):
        """Return a set of <app_label>.<permission> for this user.

        Note that these results will only be applicable to the admin panel.

        <permission> is something like add_mymodel or change_mymodel.
        """
        if hasattr(user, '_baya_cached_all_permissions'):
            return user._baya_cached_all_permissions
        permissions = super(NestedLDAPGroupsBackend, self).get_all_permissions(
            user, obj)
        from baya.admin.sites import _admin_registry
        for admin_site in _admin_registry:
            for model, opts in six.iteritems(admin_site._registry):
                app = model._meta.app_label
                perm_name = partial(get_permission_codename, opts=model._meta)
                if (hasattr(opts, 'user_has_add_permission') and
                        opts.user_has_add_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('add')))
                if (hasattr(opts, 'user_has_change_permission') and
                        opts.user_has_change_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('change')))
                if (hasattr(opts, 'user_has_delete_permission') and
                        opts.user_has_delete_permission(user)):
                    permissions.add("%s.%s" % (app, perm_name('delete')))
        user._baya_cached_all_permissions = permissions
        return permissions
